'''
Author: Pengyao PING
Date: 2022-04-01 11:36:55
LastEditors: Pengyao PING
LastEditTime: 2022-10-22 19:02:04
Email: Pengyao.Ping@student.uts.edu.au
Description: 
'''

import configparser
import os

class Config(object):
    def __init__(self, config_file, logger):
        conf = configparser.ConfigParser()
        base_path = os.getcwd()
        if config_file:
            conf.read(config_file)
            if conf.has_option("Paths", "ResultDir"):
                self.ResultDir = conf.get("Paths", "ResultDir")
            else:
                self.ResultDir = os.path.join(base_path, 'result/')

            if conf.has_option("SourceInputData", "input_file"):
                self.input_file = conf.get("SourceInputData", "input_file")
            else:
                logger.error("No input file setting in config file.")            

            if conf.has_option("SourceInputData", "ground_truth_data"):
                self.ground_truth_data = conf.get("SourceInputData", "ground_truth_data")
            else:
                self.ground_truth_data = None
                
            # setting model parameters
            # GraphSetup
            if conf.has_option("GraphSetup", "high_freq_thre"):
                self.high_freq_thre = conf.getint("GraphSetup", "high_freq_thre")
            else:
                self.high_freq_thre = 5

            if conf.has_option("GraphSetup", "max_error_freq"):
                self.max_error_freq = conf.getint("GraphSetup", "max_error_freq")
            else:
                self.max_error_freq = 3

            if conf.has_option("GraphSetup", "save_graph"):
                self.save_graph = conf.getboolean("GraphSetup", "save_graph")
            else:
                self.save_graph = False

            if conf.has_option("GraphSetup", "graph_visualization"):
                self.graph_visualization = conf.getboolean("GraphSetup", "graph_visualization")
            else:
                self.graph_visualization = False

            if conf.has_option("GraphSetup", "drawing_graph_num"):
                self.drawing_graph_num = conf.getint("GraphSetup", "drawing_graph_num")
            else:
                self.drawing_graph_num = 50   

            if conf.has_option("GraphSetup", "num_workers"):
                self.num_workers = conf.getint("GraphSetup", "num_workers")
            else:
                self.num_workers = -1 # this set will use total cpu cores - 2          

            # EmbeddingSetup
            if conf.has_option("EmbeddingSetup", "read_type"):
                self.read_type = conf.getboolean("EmbeddingSetup", "read_type")
            else:
                self.read_type = 'DNA'

            # ModelSetup
            if conf.has_option("ModelSetup", "high_ambiguous"):
                self.high_ambiguous = conf.getboolean("ModelSetup", "high_ambiguous")
            else:
                self.high_ambiguous = False        
                
            if conf.has_option("ModelTuningSetup", "n_trials"):
                self.n_trials = conf.getint("ModelTuningSetup", "n_trials")
            else:
                self.n_trials = 100
                
            if conf.has_option("ModelTuningSetup", "n_estimators"):
                self.n_estimators = conf.getint("ModelTuningSetup", "n_estimators")
            else:
                self.n_estimators = 400

            if conf.has_option("ModelTuningSetup", "test_size"):
                self.test_size = conf.getfloat("ModelTuningSetup", "test_size")
            else:
                self.test_size = 0.2 # default        
            if conf.has_option("ModelTuningSetup", "random_state"):
                self.random_state = conf.getint("ModelTuningSetup", "random_state")
            else:
                self.random_state = 2022 # default  

            if conf.has_option("ModelTuningSetup", "tree_method"):
                self.tree_method = conf.get("ModelTuningSetup", "tree_method")
            else:
                self.tree_method = 'auto'            

            if conf.has_option("ModelTuningSetup", "learning_rate_min"):
                self.learning_rate_min = conf.getfloat("ModelTuningSetup", "learning_rate_min")
            else:
                self.learning_rate_min = 1e-3 # default     
            if conf.has_option("ModelTuningSetup", "learning_rate_max"):
                self.learning_rate_max = conf.getfloat("ModelTuningSetup", "learning_rate_max")
            else:
                self.learning_rate_max = 1e-1 # default 

            if conf.has_option("ModelTuningSetup", "max_depth_min"):
                self.max_depth_min = conf.getint("ModelTuningSetup", "max_depth_min")
            else:
                self.max_depth_min = 3 # default     
            if conf.has_option("ModelTuningSetup", "max_depth_max"):
                self.max_depth_max = conf.getint("ModelTuningSetup", "max_depth_max")
            else:
                self.max_depth_max = 15 # default     
            if conf.has_option("ModelTuningSetup", "max_depth_step"):
                self.max_depth_step = conf.getint("ModelTuningSetup", "max_depth_step")
            else:
                self.max_depth_step = 1 # default 

            if conf.has_option("ModelTuningSetup", "num_boost_round_min"):
                self.num_boost_round_min = conf.getint("ModelTuningSetup", "num_boost_round_min")
            else:
                self.num_boost_round_min = 100 # default     
            if conf.has_option("ModelTuningSetup", "num_boost_round_max"):
                self.num_boost_round_max = conf.getint("ModelTuningSetup", "num_boost_round_max")
            else:
                self.num_boost_round_max = 500 # default     
            if conf.has_option("ModelTuningSetup", "num_boost_round_step"):
                self.num_boost_round_step = conf.getint("ModelTuningSetup", "num_boost_round_step")
            else:
                self.num_boost_round_step = 10 # default 

            if conf.has_option("ModelTuningSetup", "subsample_min"):
                self.subsample_min = conf.getfloat("ModelTuningSetup", "subsample_min")
            else:
                self.subsample_min = 1e-1 # default     
            if conf.has_option("ModelTuningSetup", "subsample_max"):
                self.subsample_max = conf.getfloat("ModelTuningSetup", "subsample_max")
            else:
                self.subsample_max = 99e-2 # default     

            if conf.has_option("ModelTuningSetup", "colsample_bytree_min"):
                self.colsample_bytree_min = conf.getfloat("ModelTuningSetup", "colsample_bytree_min")
            else:
                self.colsample_bytree_min = 1e-1 # default     
            if conf.has_option("ModelTuningSetup", "colsample_bytree_max"):
                self.colsample_bytree_max = conf.getfloat("ModelTuningSetup", "colsample_bytree_max")
            else:
                self.colsample_bytree_max = 99e-2 # default     
            
            if conf.has_option("ModelTuningSetup", "verbose_eval"):
                self.verbose_eval = conf.getboolean("ModelTuningSetup", "verbose_eval")
            else:
                self.verbose_eval = False
                
            # xgboostclassifier seed
            if conf.has_option("ModelTuningSetup", "seed"):
                self.seed = conf.getint("ModelTuningSetup", "seed")
            else:
                self.seed = 2022 # default 

            # if conf.has_option("ModelTuningSetup", "best_accuracy"):
            #     self.best_accuracy = conf.getfloat("ModelTuningSetup", "best_accuracy")
            # else:
            #     self.best_accuracy = 0.75 # default     

        if config_file == None:
            self.ResultDir = os.path.join(base_path, 'result/')
            self.ground_truth_data = None   
            # GraphSetup
            self.high_freq_thre = 5
            self.max_error_freq = 4
            # self.label0_min_num = 100
            self.save_graph = False
            self.graph_visualization = False
            self.drawing_graph_num = 50   
            self.num_workers = -1 # this set will use totalcpu - 2  
            # EmbeddingSetup
            self.read_type = 'DNA'
            # ModelTuningSetup
            self.high_ambiguous = False        
            self.n_trials = 10
            self.test_size = 0.1 # default        
            self.random_state = 32 # default  
            self.tree_method = 'auto'
            self.n_estimators = 400            
            self.learning_rate_min = 1e-3 # default     
            self.learning_rate_max = 1e-1 # default 
            self.max_depth_min = 3 # default     
            self.max_depth_max = 15 # default     
            self.max_depth_step = 1 # default 
            self.num_boost_round_min = 200 # default     
            self.num_boost_round_max = 300 # default     
            self.num_boost_round_step = 10 # default 
            self.subsample_min = 0.8 # default     
            self.subsample_max = 1 # default     
            self.colsample_bytree_min = 0.8 # default     
            self.colsample_bytree_max = 1 # default     
            self.verbose_eval = False
            # xgboostclassifier seed
            self.seed = 32 # default 
            self.multiprocessing_train = False
            self.best_accuracy = 0.6
            # simulation
            self.substations = True
            self.indels = False
            self.error_rate = 0.001
            # real umi
            self.umi_start = 0
            self.umi_end = 12
            self.non_umi_start = 24
            #amplicon
            self.amplicon_low_freq = 50
            self.amplicon_high_freq = 1500
            self.amplicon_threshold_proba = 0.9
            # Verbose
            self.min_iters = 100
            self.verbose = True
            # high ambiguous predict probability difference
            self.proba_deviation = 0.6
            self.iso_change_detail = True
